# -*- coding:utf-8 -*-
from ts_demuxer import TsDemux


if __name__ == '__main__':
    # 读取ts文件字节数据
    video_data = open('test.ts', 'rb').read()
    # 实例化TsDemux对象
    ts_demux = TsDemux()
    # 传递视频数据进行demux，返回ts数据（包含视频数据及音频数据）
    ts_data = ts_demux.start_demux(video_data)
    # 获取音频数据
    audio_bytes = ts_data.audio_bytes
    # 获取视频数据
    video_bytes = ts_data.video_bytes

    # 写入文件进行查看测试
    open('test_demux.mp3', 'wb').write(audio_bytes)
    open('test_demux.ts', 'wb').write(video_bytes)



# -*- coding:utf-8 -*-
from .packet import Packet
from .psi import PSI
from .pes_stream import PesStream


class TsDemux(object):
    """
    TS解复用
    """

    def __init__(self):
        self.__PSI = PSI()
        self.__PES = PesStream(self.__PSI)
        self.__CHUNK_LENGTH = 188

    def start_demux(self, video_data: bytes):
        """
        开始解复用
        :param video_data: ts流数据
        :return: video_bytes 和 audio_bytes
        """
        while len(video_data) >= self.__CHUNK_LENGTH:
            chunk = video_data[0:self.__CHUNK_LENGTH]
            packet = Packet(chunk)
            if packet.valid():
                self.__PSI.parse(packet)
                self.__PES.push_data(packet)
            video_data = video_data[188:]

        if len(video_data) < 1:
            self.__PES.assemble_one_pes()

        return TsData(self.__PES.video_bytes, self.__PES.audio_bytes)


class TsData(object):
    """
    TS音视频数据
    """

    def __init__(self, video_bytes, audio_bytes):
        self.video_bytes = video_bytes
        self.audio_bytes = audio_bytes



from .psi import PSI
from .packet import Packet
from .pes import Pes


class PesStream(object):
    def __init__(self, psi: PSI):
        self.PID = None
        self.psi = psi
        self.cache_bytes = b''
        self.audio_bytes = b''
        self.video_bytes = b''

    def push_data(self, packet: Packet):
        if 0x001f < packet.PID < 0x1fff:
            if self.psi.currentProgramPID == -1:
                self.push_packet(packet)
            elif self.psi.currentProgramPID != packet.PID:
                if packet.payload_unit_start_indicator == 1:
                    self.assemble_one_pes()
                self.push_packet(packet)

    def push_packet(self, packet: Packet):
        empty = len(self.cache_bytes) < 1
        if empty and packet.payload_unit_start_indicator == 0:
            return
        if empty:
            self.PID = packet.PID
        self.cache_bytes += packet.payload

    def assemble_one_pes(self):
        if len(self.cache_bytes) > 0:
            pes_data = Pes(self.cache_bytes)
            track = self.psi.find_track(self.PID)
            self.cache_bytes = b''
            if track:
                data = TsStream(track.id, track.stream_type, pes_data)
                if data.stream_type == 15:
                    self.audio_bytes += data.pes.data_byte
                elif data.stream_type == 27 or data.stream_type == 36:
                    self.video_bytes += data.pes.data_byte


class TsStream(object):
    def __init__(self, pid, stream_type, pes):
        self.pid = pid
        self.stream_type = stream_type
        self.pes = pes

    def __str__(self):
        return f'pid:{self.pid}\n' \
               f'stream_type:{self.stream_type}\n' \
               f'pes_length:{len(self.pes.data_byte)}'



class PATSection(object):
    def __init__(self, payload: bytes):
        section_length = ((payload[1] & 0x0f) << 8) | payload[2]
        _len = section_length - 4 - 5
        self.pmtTable = []
        for n in range(0, _len, 4):
            program_num = (payload[n + 8] << 8) | payload[n + 9]
            if program_num == 0x00:
                pass
            else:
                self.pmtTable.append(PtmTable(program_num, ((payload[10 + n] & 0x1f) << 8) | payload[11 + n]))


class PtmTable(object):
    def __init__(self, program_num, program_map_pid):
        self.programNum = program_num
        self.program_map_PID = program_map_pid



class Packet(object):
    """
    TS包
    """
    def __init__(self, _video_bytes: bytes):
        if not _video_bytes:
            raise Exception('视频数据不能为空！')

        self.sync_byte = _video_bytes[0]
        self.payload_unit_start_indicator = (_video_bytes[1] >> 6) & 1
        self.PID = ((_video_bytes[1] << 8) | _video_bytes[2]) & 0x1fff
        self.afc = (_video_bytes[3] >> 4) & 3
        self.has_payload = self.afc & 1
        self.has_adaptation = self.afc & 2
        if self.has_payload:
            if self.has_adaptation:
                adaptation_field_length = _video_bytes[4]
                self.payload = _video_bytes[5 + adaptation_field_length:]
            else:
                self.payload = _video_bytes[4:]

    def valid(self):
        sync_byte = 0x47
        return self.sync_byte == sync_byte and self.has_payload == 1

    def __str__(self):
        return f'PID:{self.PID}\nhas_payload:{self.has_payload}\n' \
               f'payload_unit_start_indicator:{self.payload_unit_start_indicator}\n' \
               f'payload:{list(self.payload)}\npayload_len:{len(self.payload)}'



from distutils.core import setup
from setuptools import find_packages

with open('README.rst', 'r', encoding='u8') as f:
    long_description = f.read()

setup(name='ts_demuxer',  # 包名
      version='1.0.0',  # 版本号
      description='ts demux',
      long_description_content_type="text/markdown",
      long_description=long_description,
      author='xuhx20',
      author_email='xuhx20@qq.com',
      url='https://gitee.com/xiao200/py_ts_demux.git',
      install_requires=[],
      license='BSD License',
      packages=find_packages(),
      platforms=["all"],
      classifiers=[
          'Intended Audience :: Developers',
          'Operating System :: OS Independent',
          'Natural Language :: Chinese (Simplified)',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries'
      ],
      )



class Pes(object):
    def __init__(self, _bytes: bytes):
        header_data_length = _bytes[8]
        self.data_byte = _bytes[9 + header_data_length:]

    def __str__(self):
        return f'data_byte:{list(self.data_byte)}'



class PMTSection(object):
    def __init__(self, payload: bytes):
        self.payload = payload
        self._get_stream_type()

    def _get_stream_type(self, ):
        section_length = ((self.payload[1] & 0x0f) << 8) | self.payload[2]
        program_info_length = ((self.payload[10] & 0x0f) << 8) | self.payload[11]
        if program_info_length < 0:
            return
        es_section_pos = 12 + program_info_length
        es_section_len = section_length - program_info_length - 9 - 4
        es_section_end = es_section_pos + es_section_len
        if es_section_pos >= es_section_end:
            print(f'es_section_pos < es_section_end {es_section_pos},{es_section_end}')
            return
        self.pes_table = []
        i = 0
        while i < es_section_len:
            base_pos = es_section_pos + i
            stream_type = self.payload[base_pos]
            elementary_pid = ((self.payload[base_pos + 1] << 8) | self.payload[base_pos + 2]) & 0x1fff
            es_info_length = ((self.payload[base_pos + 3] << 8) | self.payload[base_pos + 4]) & 0x0fff
            self.pes_table.append(PesTable(stream_type, elementary_pid))
            i += es_info_length + 5


class PesTable(object):
    def __init__(self, stream_type, pid):
        self.streamType = stream_type
        self.PID = pid



from .packet import Packet
from .pat_section import PATSection
from .pmt_section import PMTSection


class PSI(object):
    __PAT_PID = 0x0000
    __CAT_PID = 0x0001
    __TSDT_PID = 0x0002
    __SDT_PID = 0x0011

    def __init__(self):
        self.pat_table = []
        self.pes_streams = []
        self.currentProgramPID = -1

    def __getattribute__(self, attr):
        if attr == 'currentProgramPID':
            _pmtIds = []
            for i in range(len(self.pat_table)):
                _pmtIds.append(self.pat_table[i].pid)
            return _pmtIds[0] if len(_pmtIds) > 0 else -1

        return super().__getattribute__(attr)

    def parse(self, packet: Packet):
        if self.__PAT_PID == packet.PID:
            self.__parse_pat(packet)
        elif packet.PID == self.currentProgramPID:
            # pass
            self.__parse_pmt(packet)

    def __parse_pat(self, packet: Packet):
        if packet.payload_unit_start_indicator:
            pointer = packet.payload[0]
            data = packet.payload[pointer + 1:]
        else:
            data = packet.payload
        pat = PATSection(data)

        for i in range(len(pat.pmtTable)):
            self.__add_pid_to_pmt(pat.pmtTable[i].programNum, pat.pmtTable[i].program_map_PID)
        return pat

    def __parse_pmt(self, packet: Packet):
        if packet.payload_unit_start_indicator:
            pointer = packet.payload[0]
            data = packet.payload[pointer + 1:]
        else:
            data = packet.payload
        pmt = PMTSection(data)
        for i in range(len(pmt.pes_table)):
            self.__add_pes_stream(pmt.pes_table[i])
        return pmt

    def __add_pes_stream(self, pes_stream):
        def get_program(_id):
            for i in range(len(self.pes_streams)):
                if self.pes_streams[i].id == id:
                    return True
            return False

        flag = get_program(pes_stream.PID)
        if not flag:
            self.pes_streams.append(PesStreams(pes_stream.PID, pes_stream.streamType))

    def __add_pid_to_pmt(self, program_id, pid):

        def get_pmt(_id):
            for i in range(len(self.pat_table)):
                if self.pat_table[i].id == id:
                    return True
            return False

        flag = get_pmt(program_id)

        if not flag:
            self.pat_table.append(PatTable(program_id, pid))

    def find_track(self, pid):

        for i in range(len(self.pes_streams)):
            if self.pes_streams[i].id == pid:
                return self.pes_streams[i]


class PatTable(object):
    def __init__(self, _id, pid):
        self.id = _id
        self.pid = pid

    def __str__(self):
        return f'id:{self.id},pid:{self.pid}'


class PesStreams(object):
    def __init__(self, _id, stream_type):
        self.id = _id
        self.stream_type = stream_type
